# noqa
from dodo_commands.system_commands import DodoCommand
from six.moves import configparser
import os


class Command(DodoCommand):  # noqa
    safe = False
    help = "Write a value of the global Dodo Command configuration"

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('key')
        parser.add_argument('val', nargs='?')

    def _config_filename(self):
        return os.path.expanduser("~/.dodo_commands/config")

    def _write_config(self, config):
        """Save configuration."""
        with open(self._config_filename(), "w") as f:
            config.write(f)

    def _read_config(self):
        """Save configuration."""
        config = configparser.ConfigParser()
        config.read(self._config_filename())
        return config

    def handle_imp(self, key, val, **kwargs):  # noqa
        config = self._read_config()
        if val:
            config.set("DodoCommands", key, val)
            self._write_config(config)
        else:
            print(config.get("DodoCommands", key))



from dodo_commands.framework.args_tree import ArgsTreeNode

class Decorator:  # noqa
    def add_arguments(self, decorated, parser):  # noqa
        parser.add_argument(
            '--use-debugger',
            action='store_true',
            default=False,
            help="Run the command through the debugger"
        )

    def handle(self, decorated, use_debugger, **kwargs):  # noqa
        decorated.opt_use_debugger = use_debugger

    def modify_args(self, decorated, root_node, cwd):  # noqa
        if not decorated.opt_use_debugger:
            return root_node, cwd

        debugger_node = ArgsTreeNode("debugger", args=[decorated.get_config('/BUILD/debugger')])
        debugger_node.add_child(root_node)
        return debugger_node, cwd



# noqa
from dodo_commands.system_commands import DodoCommand


class Command(DodoCommand):  # noqa
    help = ""

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('what')
        parser.add_argument('where')

    def handle_imp(self, what, where, **kwargs):  # noqa
        self.runcmd(
            [
                'grep', '-rnw', where, '-e', what
            ],
        )



"""Print the full configuration."""

import ruamel.yaml
import re
from . import DodoCommand


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('--key')

    def handle_imp(self, key, **kwargs):  # noqa
        if key:
            print("%s" % str(self.get_config(key, '')))
        else:
            content = re.sub(
                r'^([0-9_A-Z]+\:)$',
                r'\n\1',
                ruamel.yaml.round_trip_dump(self.config),
                flags=re.MULTILINE
            )
            print(
                re.sub(
                    r'^\n\n',
                    r'\n',
                    content,
                    flags=re.MULTILINE
                )
            )



import plumbum
from dodo_commands.extra.standard_commands.decorators.docker import Decorator as DockerDecorator


class DiagnoseBase:
    def __init__(self, get_config, prnt, prnt_next, prnt_error):
        self.prnt = prnt
        self.prnt_next = prnt_next
        self.prnt_error = prnt_error
        self.get_config = get_config

    def _docker(self):
        try:
            return plumbum.local['docker']
        except plumbum.commands.processes.CommandNotFound:
            self.prnt_error("Docker is not installed")

    def _image(self, key_or_name):
        image = (
            self.get_config(key_or_name, None) or
            DockerDecorator.get_image_name(self.get_config, key_or_name)
        )

        if not image:
            self.prnt_error("Missing docker image configuration value for docker with name: %s" % key_or_name)
        return image

    def _print_yes_no(self, flag, yes_msg, no_msg, image):
        if flag:
            self.prnt(" yes")
            if yes_msg:
                self.prnt(yes_msg % dict(image=image))
        else:
            self.prnt(" no")
            if no_msg:
                self.prnt_next(no_msg % dict(image=image))

    def check_docker_image(self, key_or_name, yes_msg, no_msg):
        docker = self._docker()
        image = self._image(key_or_name)

        self.prnt(
            "Checking if docker image for %s (%s) is found locally:"
            % (key_or_name, image),
            new_line=False
        )

        docker_image_exists = docker("images", "-q", image)
        self._print_yes_no(docker_image_exists, yes_msg, no_msg, image)

    def check_docker_container(self, key, yes_msg, no_msg):
        docker = self._docker()
        image = self._image(key)

        self.prnt(
            "Checking if docker container ##${%s}## is found locally:" % key,
            new_line=False
        )

        container_exists = image in docker("ps", "-a", "--filter=name=%s" % image)
        self._print_yes_no(container_exists, yes_msg, no_msg, image)



"""Finds a directory or file inside the current project."""
from . import DodoCommand
from dodo_commands.framework.config import (CommandPath, get_project_dir)
import os
from six.moves import configparser
import sys


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            'what',
            nargs='?',
            help=(
                'Print the value of /ROOT/<what>_dir. For example: ' +
                '"dodo which src" prints the value of /ROOT/src_dir.')
        )

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--config',
            action="store_true",
            help='Print where the config file is')
        group.add_argument(
            '--global-config',
            action="store_true",
            help='Print where the global config file is')
        group.add_argument(
            '--script',
            help='Print where the dodo command script with given name is')
        group.add_argument(
            '--dir',
            dest='directory',
            help='Finds the X_dir value in the configuration, where X is the given option value')
        group.add_argument(
            '--decorators',
            action="store_true",
            help='Prints which command decorators are available')
        group.add_argument(
            '--projects',
            action="store_true",
            help='Prints which projects are available')

    def _which_script(self, script):
        command_path = CommandPath(self.config)
        for item in command_path.items:
            script_path = os.path.join(
                item.full_path, script + ".py"
            )
            if os.path.exists(script_path):
                return script_path
        return None

    def _which_dir(self, directory):
        return self.get_config("/ROOT/%s_dir" % directory, None)

    def _projects_dir(self):
        config = configparser.ConfigParser()
        config.read(os.path.expanduser("~/.dodo_commands/config"))
        return os.path.expanduser(config.get("DodoCommands", "projects_dir"))

    def handle_imp(self, what, config, global_config, script, directory, decorators, projects, **kwargs):
        if config:
            print(
                os.path.join(
                    get_project_dir(), "dodo_commands", "res", "config.yaml"
                )
            )
        elif global_config:
            sys.stdout.write(os.path.expanduser('~/.dodo_commands/config\n'))
        elif script:
            sys.stdout.write(self._which_script(script) + "\n")
        elif directory:
            sys.stdout.write(self._which_dir(directory) + "\n")
        elif decorators:
            sys.stdout.write(", ".join(sorted(self.all_decorators().keys())) + "\n")
        elif projects:
            projects = [
                x for x in os.listdir(self._projects_dir())
                if os.path.isdir(os.path.join(self._projects_dir(), x, "dodo_commands", "res"))
            ]
            sys.stdout.write("\n".join(sorted(projects)) + "\n")
        elif what:
            x = self._which_script(what) or self._which_dir(what)
            if x:
                sys.stdout.write(x + "\n")
        else:
            sys.stdout.write(self.get_config("/ROOT/project_name") + "\n")



"""Pauses the execution."""
import time


class Decorator:  # noqa
    def add_arguments(self, decorated, parser):  # noqa
        parser.add_argument(
            '--pause-ms',
            type=int,
            help="Pause in milliseconds before continuing"
        )

    def handle(self, decorated, pause_ms, **kwargs):  # noqa
        if pause_ms:
            time.sleep(pause_ms / 1000)

    def modify_args(self, decorated, root_node, cwd):  # noqa
        return root_node, cwd



"""Script for activating/creating a project in the projects dir."""
import argparse
from six.moves import configparser
import os
import sys
import ruamel.yaml
from plumbum import local
import dodo_commands


def _make_executable(script_filename):
    st = os.stat(script_filename)
    os.chmod(script_filename, st.st_mode | 0o111)


dodo_template = """
import os
from os.path import abspath, dirname
import sys
from dodo_commands.framework import execute_from_command_line


if __name__ == "__main__":
    if not dirname(abspath(__file__)).endswith("env/bin"):
        sys.stderr.write(
            'Error: this script must be run from the env/bin directory'
        )
        sys.exit(1)

    os.environ["DODO_COMMANDS_PROJECT_DIR"] = abspath(
        dirname(dirname(dirname(dirname(__file__))))
    )
    execute_from_command_line(sys.argv)
"""


class Activator:
    """Creates and activates projects."""

    def _config_filename(self):
        return os.path.expanduser("~/.dodo_commands/config")

    def _config(self):
        """Get configuration."""
        config = configparser.ConfigParser()
        config.read(self._config_filename())
        return config

    def _write_config(self):
        """Save configuration."""
        with open(self._config_filename(), "w") as f:
            self.config.write(f)

    def _activate_script(self):
        return os.path.join(self._dodo_commands_dir, "env", "bin", "activate")

    def _create_virtual_env(self):
        """Install a virtual env."""
        local["virtualenv"](
            "-p", self.config.get("DodoCommands", "python_interpreter"),
            os.path.join(self._dodo_commands_dir, "env")
        )

        # update activate script so that it shows the name of the
        # current project in the prompt
        with open(self._activate_script()) as f:
            lines = f.read()
        with open(self._activate_script(), "w") as f:
            f.write(lines.replace(
                r'PS1="(`basename \"$VIRTUAL_ENV\"`) $PS1"',
                r'PS1="(%s) $PS1"' % self.project
            ))

        pip = local[os.path.join(self._dodo_commands_dir, "env/bin", "pip")]
        pip(
            "install",
            "ansimarkup",
            "argcomplete",
            "parsimonious",
            "plumbum",
            "ruamel.yaml",
            "semantic_version",
            "six",
        )

        python = local[os.path.join(self._dodo_commands_dir, "env/bin", "python")]
        site_packages_dir = python(
            "-c",
            "from distutils.sysconfig import get_python_lib; " +
            "print(get_python_lib())"
        )[:-1]
        os.symlink(
            os.path.dirname(dodo_commands.__file__),
            os.path.join(site_packages_dir, "dodo_commands")
        )

    def _register_autocomplete(self):
        """Install a virtual env."""
        register = local[os.path.join(
            self._dodo_commands_dir,
            "env/bin",
            "register-python-argcomplete"
        )]
        (register >> self._activate_script())("dodo")

    def _create_dodo_script(self):
        """Install the dodo entry point script."""
        env_dir = os.path.join(self._dodo_commands_dir, "env/bin")
        dodo_file = os.path.join(env_dir, "dodo")
        with open(dodo_file, "w") as of:
            of.write('#!%s/python\n' % env_dir)
            of.write(dodo_template)
        _make_executable(dodo_file)

    def _create_res_dir(self):
        """Install the dir with dodo_commands resources."""
        res_dir = os.path.join(self._dodo_commands_dir, "res")
        os.makedirs(res_dir)
        config_filename = os.path.join(res_dir, "config.yaml")
        default_config = {
            'ROOT': {
                'command_path': [
                    ['~/.dodo_commands', 'default_commands/*']
                ],
                'version': '1.0.0'
            }
        }
        with open(config_filename, "w") as f:
            f.write(ruamel.yaml.round_trip_dump(default_config))

    def _create_project(self):
        self._report(
            "Creating project at location %s ..." % self._project_dir
        )
        self._create_res_dir()
        self._create_virtual_env()
        self._register_autocomplete()
        self._create_dodo_script()
        self._report(" done\n")
        return True

    def _report(self, x):
        sys.stderr.write(x)
        sys.stderr.flush()

    @property
    def _project_dir(self):
        return os.path.expanduser(os.path.join(
            self.config.get("DodoCommands", "projects_dir"),
            self.project
        ))

    @property
    def _dodo_commands_dir(self):
        return os.path.join(self._project_dir, "dodo_commands")

    def run(self, project, latest, create):
        """Activate or create a project in the projects dir."""
        self.project = project
        self.config = self._config()
        latest_project = (
            self.config.get("DodoCommands", "latest_project")
            if self.config.has_option("DodoCommands", "latest_project") else
            ""
        )

        if latest:
            if self.project:
                self._report(
                    "Options --latest and <project> are mutually exclusive\n"
                )
                return
            self.project = latest_project
            if not self.project:
                self._report(
                    "There is no latest project\n"
                )
                return

        if create:
            if os.path.exists(self._dodo_commands_dir):
                self._report(
                    "Project already exists: %s\n" % self._project_dir
                )
                return
            if not self._create_project():
                return
        elif not os.path.exists(self._project_dir):
            self._report(
                'Project not found: %s. Use the --create flag to create it\n'
                % self._project_dir
            )
            return

        if self.project != latest_project:
            self.config.set("DodoCommands", "latest_project", self.project)
            self._write_config()

        sys.stdout.write("source %s\n" % self._activate_script())


def main():  # noqa
    parser = argparse.ArgumentParser()
    parser.add_argument('project', nargs='?')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--create', action="store_true")
    group.add_argument('--latest', action="store_true")

    args = parser.parse_args()
    Activator().run(args.project, args.latest, args.create)



# noqa
from dodo_commands.system_commands import DodoCommand
import glob
import os
import sys
import re
from importlib import import_module
import ansimarkup
from dodo_commands.framework.util import bordered
from dodo_commands.framework.config import expand_keys
import textwrap

from ._diagnose_base import DiagnoseBase  # noqa


class Command(DodoCommand):  # noqa
    help = ""

    def add_arguments_imp(self, parser):  # noqa
        pass

    def handle_imp(self, **kwargs):  # noqa
        self.steps_dir = self.get_config('/ROOT/diagnose_dir')

        self.am = ansimarkup.AnsiMarkup(tags=dict(
            orange=ansimarkup.parse("<fg #f29130>"),
            yellow=ansimarkup.parse("<fg #faff00>")
        ))
        self._parse = self.am.parse

        step_files = glob.glob(os.path.join(self.steps_dir, "*.py"))
        for step_index in range(len(step_files) + 1):
            for step_file in step_files:
                step_module_name = os.path.splitext(
                    os.path.basename(step_file)
                )[0]
                if step_module_name.startswith("%d_" % step_index):
                    self.run_step(step_module_name)

    def _expand(self, term):
        return expand_keys(term, self.config)

    def prnt(self, msg, new_line=True):
        paragraphs = re.split('<cr/>', msg)
        for paragraph in paragraphs:
            self.prnt_paragraph(paragraph)
            if new_line:
                sys.stdout.write('\n')

    def prnt_paragraph(self, paragraph):
        msg, new_msg = paragraph, ""

        key_val_terms = re.split('\#\#([^\#]+)\#\#', msg)
        for idx, term in enumerate(key_val_terms):
            if idx % 2:
                new_msg += '<yellow>%s = %s</yellow>' % (term, self._expand(term))
            else:
                new_msg += self._expand(term)
        msg, new_msg = new_msg, ""

        yellow_terms = re.split('``([^``]+)``', msg)
        for idx, term in enumerate(yellow_terms):
            if idx % 2:
                new_msg += '<yellow>%s</yellow>' % term
            else:
                new_msg += term
        msg, new_msg = new_msg, ""

        green_terms = re.split('`([^`]+)`', msg)
        for idx, term in enumerate(green_terms):
            if idx % 2:
                new_msg += '<green>%s</green>' % term
            else:
                new_msg += term
        msg, new_msg = new_msg, ""

        sys.stdout.write(textwrap.fill(
            self._parse(msg),
            120
        ))

    def prnt_next(self, msg):
        sys.stdout.write('\n')
        self.prnt("<orange>NEXT: </orange>", new_line=False)
        self.prnt(msg)
        sys.exit(1)

    def prnt_error(self, msg):
        sys.stdout.write('\n')
        self.prnt("<red>ERROR: </red>", new_line=False)
        self.prnt(msg)
        sys.exit(1)

    def run_step(self, step_module_name):
        sys.path.append(self.steps_dir)
        module = import_module(step_module_name)
        diagnose = module.Diagnose(
            get_config=self.get_config,
            prnt=self.prnt,
            prnt_next=self.prnt_next,
            prnt_error=self.prnt_error
        )

        print(bordered(diagnose.title))
        print('')

        diagnose.run()
        print('')



# noqa
from dodo_commands.system_commands import DodoCommand, CommandError
from dodo_commands.dodo_activate import Activator


class Command(DodoCommand):  # noqa
    help = ""
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('project', nargs='?')

        group = parser.add_mutually_exclusive_group()
        group.add_argument('--latest', action="store_true")
        group.add_argument('--create', action="store_true")

    def handle_imp(self, project, latest, create, **kwargs):  # noqa
        if not project and not latest:
            raise CommandError("No project was specified")
        Activator().run(project, latest, create)



"""Enable or disable a layer in the configuration."""

from . import DodoCommand
from dodo_commands.framework.config import ConfigIO
from dodo_commands.framework import CommandError
import os


def _update_list_of_layers(layers, new_layer_name, new_layer_variant):
    def _layer_name(x):
        # layers with a prefix path are not named
        if os.path.dirname(x):
            return None
        try:
            return x.split(".")[0]
        except ValueError:
            return None

    def _new_layer():
        return "%s.%s.yaml" % (new_layer_name, new_layer_variant)

    result = []
    found = False
    for layer in layers:
        if _layer_name(layer) == new_layer_name:
            found = True
            result.append(_new_layer())
        else:
            result.append(layer)
    if not found:
        result.append(_new_layer())

    return result


def _layer_value(layers, layer):
    for l in layers:
        if l.startswith(layer + "."):
            try:
                return l.split(".")[1]
            except ValueError:
                pass
    return None


class Command(DodoCommand):  # noqa
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('layer')
        parser.add_argument('value', nargs='?', default=False)
        parser.add_argument('--force', action="store_true")

    def handle_imp(self, layer, value, force, **kwargs):  # noqa
        config = ConfigIO().load(load_layers=False)
        layers = config.get('ROOT', {}).get('layers', [])
        if value == False:  # noqa
            print(_layer_value(layers, layer))
            return

        layer_file = os.path.join(
            self.get_config("/ROOT/res_dir"),
            self.get_config("/ROOT/layer_dir", ""),
            "%s.%s.yaml" % (layer, value)
        )

        if not force and not os.path.exists(layer_file):
            raise CommandError("Layer file %s does not exist" % layer_file)

        newlayers = _update_list_of_layers(layers, layer, value)
        config['ROOT']['layers'] = newlayers
        ConfigIO().save(config)



# noqa
from dodo_commands.system_commands import DodoCommand
from dodo_commands.framework.config import expand_keys


class Command(DodoCommand):  # noqa
    help = ""

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument(
            'text'
        )

    def handle_imp(self, text, **kwargs):  # noqa
        print(expand_keys(text, self.config))



# This file is based on source code from the Django Software Foundation,
# using the following license:

# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,          # noqa
# are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.

#     3. Neither the name of Django nor the names of its contributors may be used           # noqa
#        to endorse or promote products derived from this software without
#        specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND           # noqa
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR           # noqa
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES            # noqa
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON            # noqa
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-

"""Base classes for writing management commands."""

from __future__ import unicode_literals

import os
import sys
import argcomplete
from argparse import ArgumentParser, SUPPRESS
from dodo_commands.framework.config import (
    look_up_key, load_dodo_config
)
from dodo_commands.framework.command_error import CommandError


class CommandParser(ArgumentParser):  # noqa
    """
    Customized ArgumentParser class to improve some error messages and prevent
    SystemExit in several occasions, as SystemExit is unacceptable when a
    command is called programmatically.
    """
    def __init__(self, cmd, **kwargs):  # noqa
        self.cmd = cmd
        super(CommandParser, self).__init__(**kwargs)

    def error(self, message):  # noqa
        if self.cmd._called_from_command_line:
            super(CommandParser, self).error(message)
        else:
            raise CommandError("Error: %s" % message)


class BaseCommand(object):  # noqa
    """
    The base class from which all management commands ultimately
    derive.
    """
    # Metadata about this command.
    help = ''
    args = ''

    # Configuration shortcuts that alter various logic.
    _called_from_command_line = False
    can_import_settings = True

    def __init__(self):
        self.config = load_dodo_config()

    def usage(self, subcommand):  # noqa
        """
        Return a brief description of how to use this command, by
        default from the attribute ``self.help``.
        """
        usage = '%%prog %s [options] %s' % (subcommand, self.args)
        if self.help:
            return '%s\n\n%s' % (usage, self.help)
        else:
            return usage

    def create_parser(self, prog_name, subcommand):  # noqa
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this command.
        """
        parser = CommandParser(
            self, prog="%s %s" % (os.path.basename(prog_name), subcommand),
            description=self.help or None)
        parser.add_argument(
            '--traceback', action='store_true',
            help=SUPPRESS)
        if self.args:
            # Keep compatibility and always accept positional arguments,
            # like optparse when args is set
            parser.add_argument('args', nargs='*')
        self.add_arguments(parser)
        argcomplete.autocomplete(parser)

        return parser

    def add_arguments(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        pass

    def print_help(self, prog_name, subcommand):  # noqa
        """
        Print the help message for this command, derived from
        ``self.usage()``.
        """
        parser = self.create_parser(prog_name, subcommand)
        parser.print_help()

    def run_from_argv(self, argv, subcommand):  # noqa
        """
        Set up any environment changes requested (e.g., Python path
        and Django settings), then run this command. If the
        command raises a ``CommandError``, intercept it and print it sensibly
        to stderr. If the ``--traceback`` option is present or the raised
        ``Exception`` is not ``CommandError``, raise it.
        """
        self._called_from_command_line = True
        parser = self.create_parser(argv[0], subcommand)

        options = parser.parse_args(argv[2:])

        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop('args', ())
        try:
            self.handle(*args, **cmd_options)
        except Exception as e:
            if options.traceback or not isinstance(e, CommandError):
                raise

            sys.stderr.write('%s: %s\n' % (e.__class__.__name__, e))
            sys.exit(1)

    def get_config(self, key, default_value="__not_set_234234__"):  # noqa
        return look_up_key(self.config, key, default_value)

    def handle(self, *args, **options):  # noqa
        """
        The actual logic of the command. Subclasses must implement
        this method.
        """
        raise NotImplementedError(
            'subclasses of BaseCommand must provide a handle() method')



"""Decorates command lines with docker arguments."""

from plumbum import local
from dodo_commands.framework import CommandError
from dodo_commands.framework.args_tree import ArgsTreeNode
from fnmatch import fnmatch


def _is_tuple(x):
    return hasattr(x, "__len__") and not isinstance(x, type(str()))


class Decorator:  # noqa
    @classmethod
    def _add_docker_volume_list(cls, docker_config, root_node):
        volume_list = docker_config.get('volume_list', [])
        volume_map = docker_config.get('volume_map', {})
        options = (
            ['--volume=%s:%s' % (x, x) for x in volume_list] +
            ['--volume=%s:%s' % key_val for key_val in volume_map.items()]
        )
        for x in options:
            root_node['volume'].append(x)

    @classmethod
    def _add_docker_volumes_from_list(cls, docker_config, root_node):
        volumes_from_list = docker_config.get('volumes_from_list', [])
        options = (
            ['--volumes-from=%s' % x for x in volumes_from_list]
        )
        for x in options:
            root_node['volumes-from'].append(x)

    @classmethod
    def _add_docker_variable_list(cls, docker_config, root_node):
        variable_list = docker_config.get('variable_list', [])
        variable_map = docker_config.get('variable_map', {})
        options = (
            ['--env=%s' % x for x in variable_list] +
            [
                '--env=%s=%s' % key_val
                for key_val in variable_map.items()
            ]
        )
        for x in options:
            root_node['env'].append(x)

    @classmethod
    def _add_linked_container_list(cls, docker_config, root_node):
        options = ['--link=%s' % x for x in docker_config.get('link_list', [])]
        for x in options:
            root_node['link'].append(x)

    @classmethod
    def get_image_name(cls, get_config, name):
        result = get_config('/DOCKER/options/%s/image' % name, '')
        if result:
            return result

        for patterns, docker_config in get_config('/DOCKER/options', {}).items():
            for pattern in (patterns if _is_tuple(patterns) else [patterns]):
                if name and fnmatch(name, pattern):
                    result = docker_config.get('image', None)
                    if result:
                        return result

        raise CommandError("Could not find docker image for %s" % name)

    @classmethod
    def _add_config_docker_options(cls, get_config, name, root_node):
        for key_val in get_config('/ENVIRONMENT/variable_map', {}).items():
            root_node['env'].append("--env=%s=%s" % key_val)

        for patterns, docker_config in get_config('/DOCKER/options', {}).items():
            for pattern in (patterns if _is_tuple(patterns) else [patterns]):
                if name and fnmatch(name, pattern):
                    cls._add_docker_variable_list(docker_config, root_node)
                    cls._add_docker_volume_list(docker_config, root_node)
                    cls._add_docker_volumes_from_list(docker_config, root_node)
                    cls._add_linked_container_list(docker_config, root_node)
                    for x in docker_config.get('extra_options', []):
                        root_node['other'].append(x)

    def add_arguments(self, decorated, parser):  # noqa
        parser.add_argument(
            '--non-interactive',
            action='store_true',
            help="Run docker calls without -i and -t"
        )
        parser.add_argument(
            '--kill-existing',
            action='store_true',
            help="Kill and remove existing docker container with the same name"
        )

    def handle(
        self, decorated, non_interactive, kill_existing, **kwargs
    ):  # noqa
        decorated.opt_non_interactive = non_interactive
        decorated.opt_kill_existing = kill_existing

    @classmethod
    def _add_options(cls, decorated, cwd, root_node):
        name = None

        if not hasattr(decorated, 'docker_rm') or decorated.docker_rm:
            root_node['basic'].append('--rm')
        if not decorated.opt_non_interactive:
            root_node['basic'].append('--interactive')
            root_node['basic'].append('--tty')
        if cwd:
            root_node['workdir'].append('--workdir=' + cwd)
        if hasattr(decorated, "docker_options"):
            for p in decorated.docker_options:
                try:
                    key, val = p
                    if key == "name":
                        name = val
                    child_name = key if root_node.has_child(key) else "other"
                    root_node[child_name].append(cls._to_opt(key, val))
                except:
                    raise CommandError(
                        "Elements of docker_options must be tuples of type (key, val)"
                    )

        return name

    @classmethod
    def _to_opt(cls, key, val):
        return '--%s' % key if val is None else '--%s=%s' % (key, val)

    @classmethod
    def _container_name(cls, options):
        for key, val in options:
            if key == 'name':
                return val
        return ""

    def _kill_existing_container(self, name):
        docker = local['docker']
        live_containers = docker(
            "ps",
            "--format={{.Names}}",
            "--filter=name=^/%s$" % name
        )
        if len(live_containers):
            docker("stop", name)

        exited_containers = docker(
            "ps",
            "-a",
            "--format={{.Names}}",
            "--filter=name=^/%s$" % name
        )
        if len(exited_containers):
            docker("rm", name)

    def modify_args(self, decorated, root_node, cwd):  # noqa
        docker_node = ArgsTreeNode("docker", args=['docker', 'run'])
        docker_node.add_child(ArgsTreeNode("basic"))
        docker_node.add_child(ArgsTreeNode("name"))
        docker_node.add_child(ArgsTreeNode("link"))
        docker_node.add_child(ArgsTreeNode("volume", is_horizontal=False))
        docker_node.add_child(ArgsTreeNode("volumes-from"))
        docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
        docker_node.add_child(ArgsTreeNode("other"))
        docker_node.add_child(ArgsTreeNode("workdir"))

        name = self._add_options(decorated, cwd, docker_node)
        if decorated.opt_kill_existing and name:
            self._kill_existing_container(name)

        self._add_config_docker_options(decorated.get_config, name, docker_node)

        image_name = (
            decorated.docker_image
            if hasattr(decorated, 'docker_image') else
            self.get_image_name(decorated.get_config, name)
        )
        docker_node.add_child(
            ArgsTreeNode("image", args=[image_name])
        )

        docker_node.add_child(root_node)
        return docker_node, local.cwd



"""Show diff for a file in the Dodo Commands system directory."""

from . import DodoCommand, CommandError
from six.moves import configparser
import os


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            'file',
            nargs='?',
            help='Show diff for this file'
        )
        parser.add_argument(
            '--project-name',
            help='Compare to files from an alternative project'
        )
        parser.add_argument(
            '--defaults-dir',
            help='Set the relative path to the default config files'
        )

    def _diff_tool(self):
        config = configparser.ConfigParser()
        config.read(os.path.expanduser("~/.dodo_commands/config"))
        return config.get("DodoCommands", "diff_tool")

    def handle_imp(self, file, project_name, defaults_dir, **kwargs):  # noqa
        file = file or '.'
        project_dir = self.get_config("/ROOT/project_dir", "")
        default_project_path = os.path.join(
            project_dir, "dodo_commands", "default_project"
        )

        if defaults_dir:
            if os.path.exists(default_project_path):
                raise CommandError(
                    "The --default-dir option was supplied, " +
                    "but the symlink %s already exists"
                    % default_project_path
                )

            target_dir = os.path.join(project_dir, defaults_dir)
            if not os.path.exists(target_dir):
                raise CommandError(
                    "The --default-dir option was supplied, " +
                    "but the target directory %s does not exist"
                    % target_dir
                )

            self.runcmd(["ln", "-s", target_dir, default_project_path])

        if project_name:
            ref_project_dir = os.path.abspath(
                os.path.join(project_dir, "..", project_name)
            )
            original_file = os.path.join(ref_project_dir, file)
            copied_file = os.path.join(project_dir, file)
        else:
            original_file = os.path.realpath(
                os.path.join(default_project_path, file)
            )
            copied_file = os.path.join(project_dir, "dodo_commands", "res", file)

        self.runcmd([self._diff_tool(), original_file, copied_file])



"""Finds a directory or file inside the current project."""
from dodo_commands.system_commands import DodoCommand, CommandError
from dodo_commands.framework.config import CommandPath, ConfigIO, Key, KeyNotFound
import os
from parsimonious import Grammar, ParseError
import re


grammar = Grammar(
    r"""
    api            = cwd? arg* exe_term+
    cwd            = "(" ws? term ws? ")"
    arg            = ws? "[" ws? term ws? "]"
    exe_term       = ws? (formatted_list / term)
    formatted_list = term ws "*" ws term
    term           = ~"[\%\=\-\$\{\}\/\/\._A-Z0-9]*"i
    ws             = ~"\ +"i
    """
)


class Command(DodoCommand):  # noqa
    help = "Creates a new Dodo command."
    safe = False

    @classmethod
    def _find_nodes(cls, node, expr_name, result=None):
        if result is None:
            result = []
        if node:
            if node.expr_name == expr_name:
                result.append(node)
            for child in node.children:
                cls._find_nodes(child, expr_name, result)
        return result

    @classmethod
    def _find_node(cls, node, expr_names):
        if expr_names:
            nodes = cls._find_nodes(node, expr_names[0])
            assert len(nodes) <= 1
            return cls._find_node(nodes[0], expr_names[1:]) if nodes else None
        return node

    def _args(self, args):
        if not args:
            return ""
        return ', '.join(args) + ", "

    @classmethod
    def _term(cls, term):
        def _add_quotes(x):
            return "'%s'" % x if x else None

        def _val(x):
            if (x or '').startswith('='):
                return x[1:]
            return None

        config_keys = {}
        expr = '\$\{([\/\w]+)(\=[\w\$\{\}\/]+)?\}'
        elements = re.split(expr, term)

        i = 0
        while i < len(elements):
            if i % 2 == 0:
                elements[i] = _add_quotes(elements[i])
            else:
                key, val = elements[i], elements[i + 1]
                del elements[i + 1]
                elements[i] = 'self.get_config(%s)' % _add_quotes(key)
                config_keys[key] = _val(val)
            i += 1
        return ' + '.join([x for x in elements if x]), config_keys

    def _write_arguments(self, root_node, a):
        args = []
        for arg_node in self._find_nodes(root_node, 'arg'):
            term_node = self._find_node(arg_node, ['term'])
            arg = term_node.text
            a('        parser.add_argument(\'%s\')' % arg)
            args.append(arg[2:] if arg.startswith('--') else arg)
        if not args:
            a('        pass')
        return args

    @classmethod
    def _format_list(cls, format_string, list_term, a):
        nr_args = format_string.count('%s')
        if nr_args not in [1, 2]:
            raise CommandError("Bad format string: %s" % format_string)

        a('                "%s" %% x for x in %s' % (
            format_string,
            cls._term(list_term)[0] + (".items()" if nr_args == 2 else ""),
        ))

    @classmethod
    def _write_exe_terms(cls, root_node, a, config_keys):
        a('            [')
        for exe_term_node in cls._find_nodes(root_node, 'exe_term'):
            formatted_list_node = cls._find_node(exe_term_node, ['formatted_list'])
            if formatted_list_node:
                list_term, format_string_term = cls._find_nodes(
                    formatted_list_node, 'term'
                )
                a('            ] + ')
                a('            [')
                cls._format_list(format_string_term.text, list_term.text, a)
                a('            ] + ')
                a('            [')
            else:
                term_node = cls._find_node(exe_term_node, ['term'])
                if term_node.text:
                    term, keys = cls._term(term_node.text)
                    config_keys.update(keys)
                    a('                %s,' % term)
        a('            ],')

    @classmethod
    def _write_cwd_term(cls, root_node, a, config_keys):
        cwd_node = cls._find_node(root_node, ['cwd', 'term'])
        if cwd_node:
            term, keys = cls._term(cwd_node.text)
            config_keys.update(keys)
            a('            cwd=%s' % term)

    @classmethod
    def _cast(cls, value):
        for alt_format in [int, float]:
            try:
                return alt_format(value)
            except:
                pass
        return value

    @classmethod
    def _xpath(cls, key):
        return [k for k in key.split("/") if k]

    def _has_config_key(self, xpath):
        try:
            Key(self.config, xpath).get()
            return True
        except KeyNotFound:
            return False

    def _update_config(self, config_keys, config):
        changed = False
        for key, value in config_keys.items():
            xpath = self._xpath(key)
            if self._has_config_key(xpath):
                continue

            node = config
            for idx, branch in enumerate(xpath[:-1]):
                if branch not in node:
                    node[branch] = {}
                node = node[branch]

            node[xpath[-1]] = self._cast(value)
            changed = True

        return changed

    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument('name')
        parser.add_argument(
            '--next-to',
            required=True,
            help='Create the new command at the location of this command'
        )
        parser.add_argument(
            '--parse',
            help='Parse api syntax to create the new command'
        )

    def handle_imp(self, name, next_to, parse, **kwargs):  # noqa
        dest_path = None
        command_path = CommandPath(self.config)
        for item in command_path.items:
            script_path = os.path.join(
                item.full_path, next_to + ".py"
            )
            if os.path.exists(script_path):
                dest_path = os.path.join(
                    item.full_path, name + ".py"
                )

        if not dest_path:
            raise CommandError("Script not found: %s" % next_to)

        if os.path.exists(dest_path):
            raise CommandError("Destination already exists: %s" % dest_path)

        try:
            root_node = grammar.parse(parse) if parse else None
        except ParseError as e:
            raise CommandError("Error while parsing: %s" % e)

        args = []
        config_keys = {}

        with open(dest_path, "w") as f:
            def a(x):
                f.write(x + '\n')

            a('# noqa')
            a('from dodo_commands.system_commands import DodoCommand')
            a('')
            a('')
            a('class Command(DodoCommand):  # noqa')
            a('    help = ""')
            a('')
            a('    def add_arguments_imp(self, parser):  # noqa')
            args = self._write_arguments(root_node, a)
            a('')
            a('    def handle_imp(self, %s**kwargs):  # noqa' % self._args(args))
            a('        self.runcmd(')
            self._write_exe_terms(root_node, a, config_keys)
            self._write_cwd_term(root_node, a, config_keys)
            a('        )')

        config = ConfigIO().load(load_layers=False)
        if self._update_config(config_keys, config):
            ConfigIO().save(config)

        print(dest_path)



# noqa
from dodo_commands.system_commands import DodoCommand, CommandError
import os
import sys


class Command(DodoCommand):  # noqa
    help = ""

    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            '--sudo',
            action='store_true',
        )

    def handle_imp(self, sudo, **kwargs):  # noqa
        import dodo_commands
        dodo_commands_path = dodo_commands.__path__[0]
        if os.path.realpath(dodo_commands_path) != dodo_commands_path:
            raise CommandError("Please deactivate your dodo project first by running 'deactivate'.")
        pip = os.path.join(os.path.dirname(sys.executable), "pip")

        if sudo:
            self.runcmd(["sudo", pip, "install", "--upgrade", "dodo_commands"])
        else:
            self.runcmd([pip, "install", "--upgrade", "dodo_commands"])
